#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Optional, Tuple, Union

import torch
from captum.attr import (LRP, DeepLift, DeepLiftShap, FeatureAblation,
                         FeaturePermutation, GradientShap, GuidedBackprop,
                         GuidedGradCam, InputXGradient, IntegratedGradient,
                         Lime, Occlusion, Saliency)
from torch import Tensor
from torch_geometric.data import Batch, Data
from torch_geometric.explain import Explanation
from torch_geometric.explain.algorithm.base import ExplainerAlgorithm
from torch_geometric.explain.config import (ExplainerConfig, MaskType,
                                            ModelConfig, ModelMode,
                                            ModelTaskLevel)
from torch_geometric.nn import GCNConv, global_mean_pooling
from torch_geometric.nn.models.captum import (_raise_on_invalid_mask_type,
                                              to_captum_input, to_captum_model)

import torch.nn.functional as F

class FromCaptum(ExplainerAlgorithm):
    def __init__(captum_method, mask_type: str = "node", **kwargs):
        super().__init__()
        self.captum_model = captum_model
        self.mask_type = mask_type
        _raise_on_invalid_mask_type(mask_type)
        self.kwargs = kwargs

    def supports(self) -> bool:
        task_level = self.model_config.task_level
        if task_level not in [ModelTaskLevel.graph]:
            logging.error(f"Task level '{task_level.value}' not supported")
            return False

        edge_mask_type = self.explainer_config.edge_mask_type
        if edge_mask_type not in [MaskType.object, None]:
            logging.error(f"Edge mask type '{edge_mask_type.value}' not " f"supported")
            return False

        node_mask_type = self.explainer_config.node_mask_type
        if node_mask_type not in [MaskType.attributes]:
            logging.error(f"Node mask type '{node_mask_type.value}' not " f"supported.")
            return False

        return True

    def attr_to_tuple(self, attr, mask_type):
        _raise_on_invalid_mask_type(mask_type)
        if "node" == mask_type:
            node_mask = attr
            edge_mask = None

        if "edge" == mask_type:
            node_mask = None
            edge_mask = attr

        if "node_and_mask" == mask_type:
            node_mask = attr[0]
            edge_mask = attr[1]

    def forward(
        self,
        model: torch.nn.Module,
        x: Tensor,
        edge_index: Tensor,
        target: None,
    ) -> Explanation:

        converted_model = to_captum_model(
            model,
            mask_type=self.mask_type,
            output_idx=target,
        )
        attrib = self.captum_model(converted_model)

        inputs, additional_forward_args = to_captum_input(
            x, edge_index, mask_type=self.mask_type
        )
        attr = attrib.attribute(
            inputs,
            target=target,
            additional_forward_args=additional_forward_args,
            **self.kwargs,
        )
        node_mask, edge_mask = self.attr_to_tuple(attr, self.mask_type)

        return Explanation(
            x=x, edge_index=edge_index, edge_mask=edge_mask, node_mask=node_mask
        )


if "__name__" == __main__:
    edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]], dtype=torch.long)
    x = torch.tensor([[-1], [0], [1]], dtype=torch.float)
    y = torch.tensor([1], dtype=torch.long)

    data = Data(x=x, edge_index=edge_index, y=y)
    batch = Batch().from_data_list([data])

    class Model(torch.nn.Module):
        def __init__(self, dim_in, dim_out):
            super().__init__()
            self.dim_in = dim_in
            self.dim_out = dim_out
            self.conv = GCNConv(dim_in, dim_out)

        def forward(self, batch):
            x, edge_index, batch = batch.x, batch.edge_index, batch.batch
            x, edge_index = self.conv(x, edge_index)
            x = global_mean_pooling(x, edge_index, batch)
            return x

    model = Model(1,2)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01, weight_decay=5e-4)

    for epoch in range(1, 2):
        model.train()
        optimizer.zero_grad()
        log_logits = model(batch)
        loss = F.mse_loss(out,torch.ones(2))
        loss.backward()
        optimizer.step()

    expla = [LRP, DeepLift, DeepLiftShap, FeatureAblation,
                         FeaturePermutation, GradientShap, GuidedBackprop,
                         GuidedGradCam, InputXGradient, IntegratedGradient,
                         Lime, Occlusion, Saliency]
    model = model.eval()
    for captum_exp in expla:
        try:
            exp = FromCaptum(captum_exp)
            attr = exp.forward(model,batch.x,batch.edge_index)
        except Exception as e:
            print(str(e))




    

